"""
Replace `admin.site.register` with the new `@register` decorator syntax:
https://docs.djangoproject.com/en/stable/releases/1.7/#minor-features
"""

from __future__ import annotations

import ast
from collections.abc import Iterable, MutableMapping
from functools import partial
from typing import Literal, cast
from weakref import WeakKeyDictionary

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import OP, erase_node, extract_indent, insert, reverse_find

fixer = Fixer(
    __name__,
    min_version=(1, 7),
)

# Keep track of classes that could be decorated with `@admin.register()`
# For each class name, store the associated model class names inferred from
# eligible `admin.site.register` calls.


class AdminDetails:
    __slots__ = ("parent", "lineno", "model_names_per_site")

    def __init__(self, parent: ast.AST, lineno: int) -> None:
        self.parent = parent
        self.lineno = lineno
        self.model_names_per_site: dict[str, set[str]] = {}


decorable_admins: MutableMapping[State, dict[str, AdminDetails | None]] = (
    WeakKeyDictionary()
)
# Name of site to set of unregistered model names, or True if potentially all
# models have been unregistered
unregistered_site_models: MutableMapping[State, dict[str, set[str] | Literal[True]]] = (
    WeakKeyDictionary()
)


def _is_django_admin_imported(state: State) -> bool:
    return (
        "admin" in state.from_imports["django.contrib"]
        or "admin" in state.from_imports["django.contrib.gis"]
    )


@fixer.register(ast.ClassDef)
def visit_ClassDef(
    state: State,
    node: ast.ClassDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if _is_django_admin_imported(state) and not uses_full_super_in_init_or_new(node):
        admin_detailses = decorable_admins.setdefault(state, {})
        if node.name in admin_detailses:
            # Duplicate name, ignore
            admin_detailses[node.name] = None
            return
        else:
            admin_detailses[node.name] = AdminDetails(parents[-1], node.lineno)

        if not node.decorator_list:
            offset = ast_start_offset(node)
            decorated = False
        else:
            offset = ast_start_offset(node.decorator_list[0])
            decorated = True
        yield (
            offset,
            partial(
                update_class_def,
                name=node.name,
                state=state,
                decorated=decorated,
            ),
        )


def uses_full_super_in_init_or_new(node: ast.ClassDef) -> bool:
    """
    We cannot convert classes using py2 style `super(MyAdmin, self)`
    in the `__init__` or `__new__` method.
    https://docs.djangoproject.com/en/stable/ref/contrib/admin/#the-register-decorator
    """
    visitor = FullSuperVisitor()
    visitor.generic_visit(node)
    return visitor.found_full_super


class FullSuperVisitor(ast.NodeVisitor):
    __slots__ = ("found_full_super",)

    def __init__(self) -> None:
        self.found_full_super = False

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        if node.name in ("__init__", "__new__"):
            for subnode in ast.walk(node):
                if (
                    isinstance(subnode, ast.Call)
                    and isinstance(subnode.func, ast.Name)
                    and subnode.func.id == "super"
                    and len(subnode.args) == 2
                ):
                    self.found_full_super = True

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        # Avoid descending into a new scope
        return None

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        # Avoid descending into a new scope
        return None


def update_class_def(
    tokens: list[Token], i: int, *, name: str, state: State, decorated: bool
) -> None:
    admin_details = decorable_admins.get(state, {})[name]
    if admin_details is None or not admin_details.model_names_per_site:
        return

    if decorated:
        i = reverse_find(tokens, i, name=OP, src="@")
    j, indent = extract_indent(tokens, i)

    new_src = ""
    for custom_site, model_names in sorted(admin_details.model_names_per_site.items()):
        joined_names = ", ".join(sorted(model_names))
        custom_site_src = f", site={custom_site}" if custom_site else ""
        new_src += f"{indent}@admin.register({joined_names}{custom_site_src})\n"

    insert(tokens, j, new_src=new_src)


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        _is_django_admin_imported(state)
        and isinstance(parents[-1], ast.Expr)
        and isinstance(node.func, ast.Attribute)
    ):
        if (
            node.func.attr == "register"
            and (
                (  # admin.site.register(...)
                    isinstance(node.func.value, ast.Attribute)
                    and node.func.value.attr == "site"
                    and isinstance(node.func.value.value, ast.Name)
                    and node.func.value.value.id == "admin"
                    and (site_name := "") == ""  # force assign
                )
                or (  # custom_site.register(...)
                    state.looks_like_admin_file
                    and isinstance(node.func.value, ast.Name)
                    and (site_name := node.func.value.id).endswith("site")
                )
            )
            and (
                (
                    len(node.args) == 2
                    and len(node.keywords) == 0
                    and isinstance((admin_arg := node.args[1]), ast.Name)
                )
                or (
                    len(node.args) == 1
                    and len(node.keywords) == 1
                    and node.keywords[0].arg == "admin_class"
                    and isinstance((admin_arg := node.keywords[0].value), ast.Name)
                )
            )
            and (
                isinstance((first_arg := node.args[0]), ast.Name)
                or (
                    isinstance(first_arg, (ast.Tuple, ast.List))
                    and all(isinstance(elt, ast.Name) for elt in first_arg.elts)
                )
            )
        ):
            if isinstance(first_arg, ast.Name):
                model_names = {first_arg.id}
            else:
                # cast() could be removed by using TypeGuard func above
                model_names = {cast(ast.Name, elt).id for elt in first_arg.elts}
            admin_name = admin_arg.id
            admin_details = decorable_admins.get(state, {}).get(admin_name, None)
            unregistered_models = unregistered_site_models.get(state, {}).get(
                site_name, set()
            )
            if (
                unregistered_models is not True
                and not unregistered_models.intersection(model_names)
                and admin_details is not None
                and admin_details.parent == parents[-2]
                and not (site_name and not admin_name.endswith("Admin"))
                and (
                    site_name == ""
                    or (
                        (
                            site_defined_line := get_site_defined_line(
                                parents[0], site_name
                            )
                        )
                        is not None
                        and site_defined_line < admin_details.lineno
                    )
                )
            ):
                admin_details.model_names_per_site.setdefault(site_name, set()).update(
                    model_names
                )
                yield (
                    ast_start_offset(parents[-1]),
                    partial(
                        remove_register,
                        name=admin_name,
                        state=state,
                        node=parents[-1],
                    ),
                )
        elif node.func.attr == "unregister" and (
            (  # admin.site.unregister(...)
                isinstance(node.func.value, ast.Attribute)
                and node.func.value.attr == "site"
                and isinstance(node.func.value.value, ast.Name)
                and node.func.value.value.id == "admin"
                and (site_name := "") == ""  # force assign
            )
            or (  # custom_site.unregister(...)
                state.looks_like_admin_file
                and isinstance(node.func.value, ast.Name)
                and (site_name := node.func.value.id).endswith("site")
            )
        ):
            # potentially all models unregistered, but in some cases we can
            # detect unregistered names
            unregistered_names: set[str] | Literal[True] = True
            if len(node.args) == 1:
                first_arg = node.args[0]
                if isinstance(first_arg, ast.Name):
                    unregistered_names = {first_arg.id}
                elif isinstance(first_arg, (ast.Tuple, ast.List)) and all(
                    isinstance(elt, ast.Name) for elt in first_arg.elts
                ):
                    # argument is a sequence of models
                    unregistered_names = {
                        cast(ast.Name, elt).id for elt in first_arg.elts
                    }

            state_details = unregistered_site_models.get(state, None)
            if state_details is None:
                state_details = {}
                unregistered_site_models[state] = state_details

            if unregistered_names is True:
                state_details[site_name] = True
            else:
                existing_names = state_details.get(site_name)
                if existing_names is None:
                    state_details[site_name] = unregistered_names
                elif existing_names is not True:
                    existing_names.update(unregistered_names)


def remove_register(
    tokens: list[Token], i: int, *, name: str, state: State, node: ast.Expr
) -> None:
    admin_details = decorable_admins.get(state, {})[name]
    if admin_details is None:
        return

    erase_node(tokens, i, node=node)


site_definitions: MutableMapping[ast.Module, dict[str, int | None]] = (
    WeakKeyDictionary()
)


def get_site_defined_line(module: ast.AST, site_name: str) -> int | None:
    assert isinstance(module, ast.Module)
    lines = site_definitions.get(module, None)
    if lines is None:
        lines = {}
        for node in module.body:
            if isinstance(node, ast.ImportFrom):
                for alias in node.names:
                    if alias.asname is not None:
                        name = alias.asname
                    else:
                        name = alias.name

                    if name.endswith("site") and name not in lines:
                        lines[name] = node.lineno
            elif (
                isinstance(node, ast.Assign)
                and len(node.targets) == 1
                and isinstance(node.targets[0], ast.Name)
                and (name := node.targets[0].id) not in lines
            ):
                lines[name] = node.lineno

        site_definitions[module] = lines
    return lines.get(site_name)



"""
Replace BaseCommand.requires_system_checks boolean flag by list of checks:
https://docs.djangoproject.com/en/stable/releases/3.2/#deprecated-features-3-2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import replace

fixer = Fixer(
    __name__,
    min_version=(3, 2),
    condition=lambda state: state.looks_like_command_file,
)


@fixer.register(ast.Assign)
def visit_Assign(
    state: State,
    node: ast.Assign,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(parents[-1], ast.ClassDef)
        and len(node.targets) == 1
        and isinstance(node.targets[0], ast.Name)
        and node.targets[0].id == "requires_system_checks"
        and isinstance(node.value, ast.Constant)
        and (node.value.value is True or node.value.value is False)
    ):
        if node.value.value:
            new_src = '"__all__"'
        else:
            new_src = "[]"
        yield ast_start_offset(node.value), partial(replace, src=new_src)



"""
Replace declarations of database support in test cases:
https://docs.djangoproject.com/en/2.2/releases/2.2/#features-deprecated-in-2-2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import CODE, find_last_token

fixer = Fixer(
    __name__,
    min_version=(2, 2),
    condition=lambda state: state.looks_like_test_file,
)


@fixer.register(ast.Assign)
def visit_Assign(
    state: State,
    node: ast.Assign,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(parents[-1], ast.ClassDef)
        and len(node.targets) == 1
        and isinstance(node.targets[0], ast.Name)
        and node.targets[0].id in ("allow_database_queries", "multi_db")
        and isinstance(node.value, ast.Constant)
        and (node.value.value is True or node.value.value is False)
    ):
        yield (
            ast_start_offset(node),
            partial(replace_assignment, node=node, value=node.value.value),
        )


def replace_assignment(
    tokens: list[Token], i: int, *, node: ast.Assign, value: bool
) -> None:
    new_src = "databases = "
    if value:
        new_src += '"__all__"'
    else:
        new_src += "[]"
    j = find_last_token(tokens, i, node=node)
    tokens[i : j + 1] = [Token(name=CODE, src=new_src)]



"""
Rewrite CheckConstraint calls to use 'condition' argument instead of 'check':
https://docs.djangoproject.com/en/5.1/releases/5.1/#id2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import replace

fixer = Fixer(
    __name__,
    min_version=(5, 1),
)


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            (
                isinstance(node.func, ast.Name)
                and node.func.id == "CheckConstraint"
                and (
                    "CheckConstraint" in state.from_imports["django.db.models"]
                    or "CheckConstraint"
                    in state.from_imports["django.contrib.gis.db.models"]
                )
            )
            or (
                isinstance(node.func, ast.Attribute)
                and node.func.attr == "CheckConstraint"
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "models"
                and (
                    "models" in state.from_imports["django.db"]
                    or "models" in state.from_imports["django.contrib.gis.db"]
                )
            )
        )
        and (kwarg_names := {k.arg for k in node.keywords})
        and "check" in kwarg_names
        and "condition" not in kwarg_names
    ):
        check_kwarg = [k for k in node.keywords if k.arg == "check"][0]
        yield ast_start_offset(check_kwarg), partial(replace, src="condition")



"""
Rewrite django.core.paginator’s alias QuerySetPaginator of Paginator:
https://docs.djangoproject.com/en/2.2/releases/2.2/#features-deprecated-in-2-2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import find_and_replace_name, update_import_names

fixer = Fixer(
    __name__,
    min_version=(2, 2),
)

MODULE = "django.core.paginator"
NAMES = {
    "QuerySetPaginator": "Paginator",
}


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if node.module == MODULE and is_rewritable_import_from(node):
        yield (
            ast_start_offset(node),
            partial(update_import_names, node=node, name_map=NAMES),
        )


@fixer.register(ast.Name)
def visit_Name(
    state: State,
    node: ast.Name,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (name := node.id) in NAMES and name in state.from_imports[MODULE]:
        yield (
            ast_start_offset(node),
            partial(find_and_replace_name, name=name, new=NAMES[name]),
        )


@fixer.register(ast.Attribute)
def visit_Attribute(
    state: State,
    node: ast.Attribute,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    name = node.attr
    if (
        name in NAMES
        and isinstance(node.value, ast.Name)
        and node.value.id == "paginator"
        and "paginator" in state.from_imports["django.core"]
    ):
        yield (
            ast_start_offset(node),
            partial(find_and_replace_name, name=name, new=NAMES[name]),
        )



"""
Rewrite some format_html() calls passing formatted strings without other
arguments or keyword arguments to use the format_html formatting.

https://docs.djangoproject.com/en/5.0/releases/5.0/#features-deprecated-in-5-0
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import OP, alone_on_line, find, find_last_token, insert

fixer = Fixer(
    __name__,
    min_version=(5, 0),
)


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        "format_html" in state.from_imports["django.utils.html"]
        and isinstance(node.func, ast.Name)
        and node.func.id == "format_html"
        # Template only
        and len(node.args) == 1
        and len(node.keywords) == 0
        # str.format()
        and isinstance((str_format := node.args[0]), ast.Call)
        and isinstance(str_format.func, ast.Attribute)
        and isinstance(str_format.func.value, ast.Constant)
        and isinstance(str_format.func.value.value, str)
        and str_format.func.attr == "format"
    ):
        yield ast_start_offset(node), partial(rewrite_str_format, node=str_format)


def rewrite_str_format(
    tokens: list[Token],
    i: int,
    *,
    node: ast.Call,
) -> None:
    open_start = find(tokens, i, name=OP, src=".")
    open_end = find(tokens, open_start, name=OP, src="(")

    # closing paren
    cp_start = cp_end = find_last_token(tokens, open_end, node=node)
    if alone_on_line(tokens, cp_start, cp_end):
        cp_start -= 1
        cp_end += 1

    del tokens[cp_start : cp_end + 1]
    del tokens[open_start : open_end + 1]
    insert(tokens, open_start, new_src=", ")



"""
Update use of request.META to fetch headers to use request.headers
https://docs.djangoproject.com/en/2.2/releases/2.2/#requests-and-responses
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import NAME, STRING, find, replace, str_repr_matching

fixer = Fixer(
    __name__,
    min_version=(2, 2),
)


@fixer.register(ast.Subscript)
def visit_Subscript(
    state: State,
    node: ast.Subscript,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        is_request_or_self_request_meta(node.value)
        and not isinstance(parents[-1], ast.Delete)
        and not (isinstance(parents[-1], ast.Assign) and node in parents[-1].targets)
        and (meta_name := extract_constant(node.slice)) is not None
        and (header_name := get_header_name(meta_name)) is not None
    ):
        yield (
            ast_start_offset(node),
            partial(rewrite_header_access, header_name=header_name),
        )


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(node.func, ast.Attribute)
        and node.func.attr == "get"
        and is_request_or_self_request_meta(node.func.value)
        and len(node.args) >= 1
        and isinstance(node.args[0], ast.Constant)
        and isinstance(meta_name := node.args[0].value, str)
        and (header_name := get_header_name(meta_name)) is not None
    ):
        yield (
            ast_start_offset(node),
            partial(rewrite_header_access, header_name=header_name),
        )


@fixer.register(ast.Compare)
def visit_Compare(
    state: State,
    node: ast.Compare,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        len(node.ops) == 1
        and isinstance(node.ops[0], (ast.In, ast.NotIn))
        and len(node.comparators) == 1
        and is_request_or_self_request_meta(node.comparators[0])
        and isinstance(node.left, ast.Constant)
        and isinstance(node.left.value, str)
        and (header_name := get_header_name(node.left.value)) is not None
    ):
        yield (
            ast_start_offset(node),
            partial(rewrite_in_statement, header_name=header_name),
        )


def is_request_or_self_request_meta(node: ast.AST) -> bool:
    return (
        isinstance(node, ast.Attribute)
        and node.attr == "META"
        and (
            (isinstance(node.value, ast.Name) and node.value.id == "request")
            or (
                isinstance(node.value, ast.Attribute)
                and isinstance(node.value.value, ast.Name)
                and node.value.value.id == "self"
                and node.value.attr == "request"
            )
        )
    )


def extract_constant(node: ast.AST) -> str | None:
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        return node.value
    return None


def get_header_name(meta_name: str) -> str | None:
    """Extract HTTP header name, unless it isn't an HTTP header."""
    http_prefix = "HTTP_"
    if meta_name.startswith(http_prefix):
        name = meta_name[len(http_prefix) :]
    elif meta_name in {"CONTENT_LENGTH", "CONTENT_TYPE"}:
        name = meta_name
    else:
        return None
    return "-".join(x for x in name.lower().split("_"))


def rewrite_header_access(tokens: list[Token], i: int, *, header_name: str) -> None:
    meta_idx = find(tokens, i, name=NAME, src="META")
    replace(tokens, meta_idx, src="headers")
    str_idx = find(tokens, meta_idx, name=STRING)
    header_src = str_repr_matching(header_name, match_quotes=tokens[str_idx].src)
    replace(tokens, str_idx, src=header_src)


def rewrite_in_statement(tokens: list[Token], i: int, *, header_name: str) -> None:
    header_src = str_repr_matching(header_name, match_quotes=tokens[i].src)
    replace(tokens, i, src=header_src)
    meta_idx = find(tokens, i, name=NAME, src="META")
    replace(tokens, meta_idx, src="headers")



from __future__ import annotations

import ast
import pkgutil
import re
from collections import defaultdict
from collections.abc import Iterable
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from tokenize_rt import Offset, Token

from django_upgrade import fixers


class Settings:
    __slots__ = (
        "target_version",
        "enabled_fixers",
    )

    def __init__(
        self,
        target_version: tuple[int, int],
        only_fixers: set[str] | None = None,
        skip_fixers: set[str] | None = None,
    ) -> None:
        self.target_version = target_version
        self.enabled_fixers = {
            name
            for name in FIXERS
            if (only_fixers is None or name in only_fixers)
            and (skip_fixers is None or name not in skip_fixers)
        }


apps_re = re.compile(r"(^|[\\/])apps\.py$")
admin_re = re.compile(r"(\b|_)admin(\b|_)")
commands_re = re.compile(r"(^|[\\/])management[\\/]commands[\\/]")
dunder_init_re = re.compile(r"(^|[\\/])__init__\.py$")
migrations_re = re.compile(r"(^|[\\/])migrations([\\/])")
settings_re = re.compile(r"(\b|_)settings(\b|_)")
test_re = re.compile(r"(\b|_)tests?(\b|_)")
models_re = re.compile(r"(^|[\\/])models([\\/]|\.py)")


class State:
    __slots__ = ("settings", "filename", "from_imports", "__weakref__", "__dict__")

    def __init__(
        self,
        settings: Settings,
        filename: str,
        from_imports: defaultdict[str, set[str]],
    ) -> None:
        self.settings = settings
        self.filename = filename
        self.from_imports = from_imports

    @cached_property
    def looks_like_admin_file(self) -> bool:
        return admin_re.search(self.filename) is not None

    @cached_property
    def looks_like_command_file(self) -> bool:
        return commands_re.search(self.filename) is not None

    @cached_property
    def looks_like_dunder_init_file(self) -> bool:
        return dunder_init_re.search(self.filename) is not None

    @cached_property
    def looks_like_migrations_file(self) -> bool:
        return migrations_re.search(self.filename) is not None

    @cached_property
    def looks_like_settings_file(self) -> bool:
        return settings_re.search(self.filename) is not None

    @cached_property
    def looks_like_test_file(self) -> bool:
        return test_re.search(self.filename) is not None

    @cached_property
    def looks_like_models_file(self) -> bool:
        return models_re.search(self.filename) is not None


AST_T = TypeVar("AST_T", bound=ast.AST)
TokenFunc = Callable[[list[Token], int], None]
ASTFunc = Callable[
    [State, AST_T, tuple[ast.AST, ...]], Iterable[tuple[Offset, TokenFunc]]
]

if TYPE_CHECKING:  # pragma: no cover
    from typing import Protocol
else:
    Protocol = object


class ASTCallbackMapping(Protocol):
    def __getitem__(self, tp: type[AST_T]) -> list[ASTFunc[AST_T]]:  # pragma: no cover
        ...

    def items(self) -> Iterable[tuple[Any, Any]]:  # pragma: no cover
        ...


def visit(
    tree: ast.Module,
    settings: Settings,
    filename: str,
) -> dict[Offset, list[TokenFunc]]:
    state = State(
        settings=settings,
        filename=filename,
        from_imports=defaultdict(set),
    )
    ast_funcs = get_ast_funcs(state, settings)

    nodes: list[tuple[ast.AST, tuple[ast.AST, ...]]] = [(tree, ())]
    ret = defaultdict(list)
    while nodes:
        node, parents = nodes.pop()

        for ast_func in ast_funcs[type(node)]:
            for offset, token_func in ast_func(state, node, parents):
                ret[offset].append(token_func)

        if (
            isinstance(node, ast.ImportFrom)
            and node.level == 0
            and (
                node.module is not None
                and (
                    node.module.startswith("django.")
                    or node.module in ("django", "unittest")
                )
            )
        ):
            state.from_imports[node.module].update(
                name.name
                for name in node.names
                if name.asname is None and name.name != "*"
            )

        subparents = parents + (node,)
        for name in reversed(node._fields):
            value = getattr(node, name)

            if isinstance(value, ast.AST):
                nodes.append((value, subparents))
            elif isinstance(value, list):
                for subvalue in reversed(value):
                    if isinstance(subvalue, ast.AST):
                        nodes.append((subvalue, subparents))
    return ret


class Fixer:
    __slots__ = (
        "name",
        "min_version",
        "ast_funcs",
        "condition",
    )

    def __init__(
        self,
        module_name: str,
        min_version: tuple[int, int],
        condition: Callable[[State], bool] | None = None,
    ) -> None:
        self.name = module_name.rpartition(".")[2]
        self.min_version = min_version
        self.ast_funcs: ASTCallbackMapping = defaultdict(list)
        self.condition = condition

        FIXERS[self.name] = self

    def register(
        self, type_: type[AST_T]
    ) -> Callable[[ASTFunc[AST_T]], ASTFunc[AST_T]]:
        def decorator(func: ASTFunc[AST_T]) -> ASTFunc[AST_T]:
            self.ast_funcs[type_].append(func)
            return func

        return decorator


FIXERS: dict[str, Fixer] = {}


def _import_fixers() -> None:
    # https://github.com/python/mypy/issues/1422
    fixers_path: str = fixers.__path__  # type: ignore [assignment]
    mod_infos = pkgutil.walk_packages(fixers_path, f"{fixers.__name__}.")
    for _, name, _ in mod_infos:
        __import__(name, fromlist=["_trash"])


_import_fixers()


def get_ast_funcs(state: State, settings: Settings) -> ASTCallbackMapping:
    ast_funcs: ASTCallbackMapping = defaultdict(list)
    for fixer in FIXERS.values():
        if fixer.name not in settings.enabled_fixers:
            continue
        if fixer.min_version <= state.settings.target_version and (
            fixer.condition is None or fixer.condition(state)
        ):
            for type_, type_funcs in fixer.ast_funcs.items():
                ast_funcs[type_].extend(type_funcs)
    return ast_funcs



"""
Rewrite django.core.validator.EmailValidator arguments:
https://docs.djangoproject.com/en/3.2/releases/3.2/#features-deprecated-in-3-2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import replace_argument_names

fixer = Fixer(
    __name__,
    min_version=(3, 2),
)

MODULE = "django.core.validators"
NAME = "EmailValidator"
KWARGS = {"whitelist": "allowlist"}


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            isinstance(node.func, ast.Name)
            and NAME in state.from_imports[MODULE]
            and node.func.id == NAME
        )
        or (
            isinstance(node.func, ast.Attribute)
            and node.func.attr == NAME
            and "validators" in state.from_imports["django.core"]
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "validators"
        )
    ) and any(k.arg in KWARGS for k in node.keywords):
        yield (
            ast_start_offset(node),
            partial(
                replace_argument_names,
                node=node,
                arg_map=KWARGS,
            ),
        )



"""
Rename django.contrib.admin.utils.lookup_needs_distinct to lookup_spawns_duplicates:
https://docs.djangoproject.com/en/4.0/releases/4.0/#miscellaneous
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import find_and_replace_name, update_import_names

fixer = Fixer(
    __name__,
    min_version=(4, 0),
)

MODULE = "django.contrib.admin.utils"
RENAMES = {
    "lookup_needs_distinct": "lookup_spawns_duplicates",
}


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if node.module == MODULE and is_rewritable_import_from(node):
        name_map = {}
        for alias in node.names:
            if alias.name in RENAMES:
                name_map[alias.name] = RENAMES[alias.name]

        if name_map:
            yield (
                ast_start_offset(node),
                partial(
                    update_import_names,
                    node=node,
                    name_map=name_map,
                ),
            )


@fixer.register(ast.Name)
def visit_Name(
    state: State,
    node: ast.Name,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (name := node.id) in RENAMES and name in state.from_imports[MODULE]:
        new_name = RENAMES[name]
        yield (
            ast_start_offset(node),
            partial(find_and_replace_name, name=name, new=new_name),
        )



"""
Update use of request.user methods is_anonymous() and is_authenticated() to
attribute accesses.

https://docs.djangoproject.com/en/1.10/releases/1.10/#using-user-is-authenticated-and-user-is-anonymous-as-methods
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import NAME, OP, find

fixer = Fixer(
    __name__,
    min_version=(1, 10),
)


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(node.func, ast.Attribute)
        and node.func.attr in ("is_anonymous", "is_authenticated")
        and len(node.args) == 0
        and is_request_or_self_request_user(node.func.value)
    ):
        yield (
            ast_start_offset(node),
            partial(rewrite_user_attribute, attr=node.func.attr),
        )


def is_request_or_self_request_user(node: ast.AST) -> bool:
    return (
        isinstance(node, ast.Attribute)
        and node.attr == "user"
        and (
            (isinstance(node.value, ast.Name) and node.value.id == "request")
            or (
                isinstance(node.value, ast.Attribute)
                and isinstance(node.value.value, ast.Name)
                and node.value.value.id == "self"
                and node.value.attr == "request"
            )
        )
    )


def rewrite_user_attribute(tokens: list[Token], i: int, *, attr: str) -> None:
    j = find(tokens, i, name=NAME, src=attr)
    y = find(tokens, j, name=OP, src="(")
    z = find(tokens, y, name=OP, src=")")
    del tokens[y : z + 1]



"""
Drop test skip decorators for old Django versions like:

import unittest

import django
import pytest
from django.test import TestCase

class ExampleTests(TestCase):
    @unittest.skipIf(django.VERSION < (5, 1), "Django 5.1+")
    def test_one(self):
        ...

    @unittest.skipUnless(django.VERSION >= (5, 1), "Django 5.1+")
    def test_two(self):
        ...

    @pytest.mark.skipif(django.VERSION < (5, 1), reason="Django 5.1+")
    def test_three(self):
        ...
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset, is_passing_comparison
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import erase_decorator

fixer = Fixer(
    __name__,
    min_version=(0, 0),
)


@fixer.register(ast.FunctionDef)
def visit_FunctionDef(
    state: State,
    node: ast.FunctionDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    yield from _handle_decorator(state, node, parents)


@fixer.register(ast.ClassDef)
def visit_ClassDef(
    state: State,
    node: ast.ClassDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    yield from _handle_decorator(state, node, parents)


def _handle_decorator(
    state: State,
    node: ast.FunctionDef | ast.ClassDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    for decorator in node.decorator_list:
        if (
            isinstance(decorator, ast.Call)
            and (
                (
                    isinstance(decorator.func, ast.Attribute)
                    and isinstance(decorator.func.value, ast.Name)
                    and decorator.func.value.id == "unittest"
                    and decorator.func.attr in ("skipIf", "skipUnless")
                    and (ident := ("unittest", decorator.func.attr))
                )
                or (
                    isinstance(decorator.func, ast.Name)
                    and (decorator.func.id in ("skipIf", "skipUnless"))
                    and decorator.func.id in state.from_imports["unittest"]
                    and (ident := ("unittest", decorator.func.id))
                )
                # or pytest.mark.skipif
                or (
                    isinstance(decorator.func, ast.Attribute)
                    and isinstance(decorator.func.value, ast.Attribute)
                    and isinstance(decorator.func.value.value, ast.Name)
                    and decorator.func.value.value.id == "pytest"
                    and decorator.func.value.attr == "mark"
                    and decorator.func.attr == "skipif"
                    and (ident := ("pytest", "mark.skipif"))
                )
            )
            and (
                (
                    ident[0] == "unittest"
                    and len(decorator.args) == 2
                    and len(decorator.keywords) == 0
                )
                or (
                    ident[0] == "pytest"
                    and len(decorator.args) == 1
                    and len(decorator.keywords) == 1
                    and decorator.keywords[0].arg == "reason"
                )
            )
            and isinstance(decorator.args[0], ast.Compare)
            and (
                (pass_fail := is_passing_comparison(decorator.args[0], state))
                is not None
            )
            and (
                (ident == ("unittest", "skipIf") and pass_fail == "fail")
                or (ident == ("unittest", "skipUnless") and pass_fail == "pass")
                or (ident == ("pytest", "mark.skipif") and pass_fail == "fail")
            )
        ):
            yield ast_start_offset(decorator), partial(erase_decorator, node=decorator)



"""
Replace `list` error message key with `list_invalid` for ModelMultipleChoiceField.
https://docs.djangoproject.com/en/3.1/releases/3.1/#id2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import replace

fixer = Fixer(
    __name__,
    min_version=(3, 1),
)


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            (
                isinstance(node.func, ast.Attribute)
                and node.func.attr == "ModelMultipleChoiceField"
                and "forms" in state.from_imports["django"]
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "forms"
            )
            or (
                isinstance(node.func, ast.Name)
                and node.func.id == "ModelMultipleChoiceField"
                and node.func.id in state.from_imports["django.forms"]
            )
        )
        and any(
            (error_message_node := kw).arg == "error_messages" for kw in node.keywords
        )
        and isinstance(error_message_node.value, ast.Dict)
        and any(
            (isinstance(key, ast.Constant) and (list_node := key).value == "list")
            for key in error_message_node.value.keys
        )
    ):
        yield ast_start_offset(list_node), partial(replace, src='"invalid_list"')



"""
Convert positional arguments to keyword arguments for Django email APIs:
https://docs.djangoproject.com/en/6.0/releases/6.0/#positional-arguments-in-django-core-mail-apis
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import CODE, OP, find, parse_call_args

fixer = Fixer(
    __name__,
    min_version=(6, 0),
)


class APIConfig:
    __slots__ = ("new_posargs", "new_kwargs", "extra_kwargs")

    def __init__(
        self, new_posargs: int, new_kwargs: list[str], extra_kwargs: bool
    ) -> None:
        self.new_posargs = new_posargs
        self.new_kwargs = new_kwargs
        self.extra_kwargs = extra_kwargs


API_CONFIGS = {
    "get_connection": APIConfig(1, ["fail_silently"], True),
    "mail_admins": APIConfig(2, ["fail_silently", "connection", "html_message"], False),
    "mail_managers": APIConfig(
        2, ["fail_silently", "connection", "html_message"], False
    ),
    "send_mail": APIConfig(
        4,
        [
            "fail_silently",
            "auth_user",
            "auth_password",
            "connection",
            "html_message",
        ],
        False,
    ),
    "send_mass_mail": APIConfig(
        1, ["fail_silently", "auth_user", "auth_password", "connection"], False
    ),
    "EmailMessage": APIConfig(
        4,
        [
            "bcc",
            "connection",
            "attachments",
            "headers",
            "cc",
            "reply_to",
        ],
        False,
    ),
    "EmailMultiAlternatives": APIConfig(
        4,
        [
            "bcc",
            "connection",
            "attachments",
            "headers",
            "alternatives",
            "cc",
            "reply_to",
        ],
        False,
    ),
}

MESSAGE_MODULE_NAMES = {"EmailMessage", "EmailMultiAlternatives"}


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    # Check for direct import or module import and get function config
    if (
        isinstance(node.func, ast.Name)
        and (
            (
                (func_name := node.func.id) in API_CONFIGS
                and func_name in state.from_imports["django.core.mail"]
            )
            or (
                func_name in MESSAGE_MODULE_NAMES
                and func_name in state.from_imports["django.core.mail.message"]
            )
        )
    ) or (
        isinstance(node.func, ast.Attribute)
        and (func_name := node.func.attr) in API_CONFIGS
        and isinstance(node.func.value, ast.Name)
        and (
            (
                node.func.value.id == "mail"
                and "mail" in state.from_imports["django.core"]
            )
            or (
                func_name in MESSAGE_MODULE_NAMES
                and node.func.value.id == "message"
                and "message" in state.from_imports["django.core.mail"]
            )
        )
    ):
        api_config = API_CONFIGS[func_name]
        num_posargs = len(node.args)
        convertible_posargs = num_posargs - api_config.new_posargs

        if convertible_posargs <= 0 or convertible_posargs > len(api_config.new_kwargs):
            return

        convertible_kwargs = api_config.new_kwargs[:convertible_posargs]

        existing_kwargs = {kw.arg for kw in node.keywords}
        if any(kw in existing_kwargs for kw in convertible_kwargs):
            return

        if not api_config.extra_kwargs:
            unknown_kwargs = existing_kwargs - set(api_config.new_kwargs)
            if unknown_kwargs:
                return

        yield (
            ast_start_offset(node),
            partial(
                migrate_api_args,
                api_config=api_config,
                num_posargs=num_posargs,
                convertible_posargs=convertible_posargs,
            ),
        )


def migrate_api_args(
    tokens: list[Token],
    i: int,
    *,
    api_config: APIConfig,
    num_posargs: int,
    convertible_posargs: int,
) -> None:
    """
    Convert excess positional arguments to keyword arguments for email functions.
    Only converts arguments beyond the allowed positional count.
    """
    open_idx = find(tokens, i, name=OP, src="(")
    func_args, _close_idx = parse_call_args(tokens, open_idx)

    for argindex in range(num_posargs - 1, num_posargs - convertible_posargs - 1, -1):
        kwarg = api_config.new_kwargs[argindex - api_config.new_posargs]
        arg_start, arg_end = func_args[argindex]

        actual_arg_start = arg_start
        while actual_arg_start < arg_end and tokens[actual_arg_start].name in (
            "UNIMPORTANT_WS",
            "NL",
        ):
            actual_arg_start += 1
        tokens.insert(actual_arg_start, Token(name=CODE, src=f"{kwarg}="))



"""
Replace imports from django.utils.http:
https://docs.djangoproject.com/en/3.0/releases/3.0/#features-deprecated-in-3-0
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    extract_indent,
    find_and_replace_name,
    insert,
    update_import_names,
)

fixer = Fixer(
    __name__,
    min_version=(3, 0),
)

MODULE = "django.utils.http"
RENAMES = {
    "is_safe_url": "url_has_allowed_host_and_scheme",
}
URLLIB_NAMES = {
    "urlquote": "quote",
    "urlquote_plus": "quote_plus",
    "urlunquote": "unquote",
    "urlunquote_plus": "unquote_plus",
}


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if node.module == MODULE and is_rewritable_import_from(node):
        name_map = {}
        urllib_names = {}
        for alias in node.names:
            if alias.name in RENAMES:
                name_map[alias.name] = RENAMES[alias.name]
            elif alias.name in URLLIB_NAMES:
                name_map[alias.name] = ""
                urllib_names[alias.name] = alias.asname

        if name_map:
            yield (
                ast_start_offset(node),
                partial(
                    fix_import,
                    node=node,
                    name_map=name_map,
                    urllib_names=urllib_names,
                ),
            )


def fix_import(
    tokens: list[Token],
    i: int,
    *,
    node: ast.ImportFrom,
    name_map: dict[str, str],
    urllib_names: dict[str, str | None],
) -> None:
    j, indent = extract_indent(tokens, i)

    update_import_names(tokens, i, node=node, name_map=name_map)

    if urllib_names:
        urllib_imports = []
        for name, asname in urllib_names.items():
            if asname is None:
                urllib_imports.append(URLLIB_NAMES[name])
            else:
                urllib_imports.append(f"{URLLIB_NAMES[name]} as {asname}")

        insert(
            tokens,
            j,
            new_src=f"{indent}from urllib.parse import {', '.join(urllib_imports)}\n",
        )


@fixer.register(ast.Name)
def visit_Name(
    state: State,
    node: ast.Name,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (name := node.id) in state.from_imports[MODULE]:
        new_name: str | None
        if name in RENAMES:
            new_name = RENAMES[name]
        elif name in URLLIB_NAMES:
            new_name = URLLIB_NAMES[name]
        else:
            new_name = None

        if new_name is not None:
            yield (
                ast_start_offset(node),
                partial(find_and_replace_name, name=name, new=new_name),
            )



"""
Replace django.utils.timezone.utc with datetime.timezone.utc
https://docs.djangoproject.com/en/4.1/releases/4.1/#id2
"""

from __future__ import annotations

import ast
from collections.abc import Iterable, MutableMapping
from functools import partial
from weakref import WeakKeyDictionary

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import replace, update_import_names

fixer = Fixer(
    __name__,
    min_version=(4, 1),
)


@fixer.register(ast.Name)
def visit_Name(
    state: State,
    node: ast.Name,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        node.id == "utc"
        and (
            (details := get_import_details(state, parents[0])).old_utc_import
            is not None
        )
        and details.datetime_module is not None
    ):
        yield from maybe_rewrite_import(details)
        new_src = f"{details.datetime_module}.timezone.utc"
        yield ast_start_offset(node), partial(replace, src=new_src)


@fixer.register(ast.Attribute)
def visit_Attribute(
    state: State,
    node: ast.Attribute,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        node.attr == "utc"
        and isinstance(node.value, ast.Name)
        and node.value.id == "timezone"
        and "timezone" in state.from_imports["django.utils"]
        and (details := get_import_details(state, parents[0])).datetime_module
        is not None
    ):
        new_src = f"{details.datetime_module}.timezone"
        yield ast_start_offset(node), partial(replace, src=new_src)


class ImportDetails:
    __slots__ = (
        "old_utc_import",
        "datetime_module",
        "rewrite_scheduled",
    )

    def __init__(self) -> None:
        self.old_utc_import: ast.ImportFrom | None = None
        self.datetime_module: str | None = None
        self.rewrite_scheduled = False


import_details: MutableMapping[State, ImportDetails] = WeakKeyDictionary()


def get_import_details(state: State, module: ast.AST) -> ImportDetails:
    assert isinstance(module, ast.Module)
    try:
        return import_details[state]
    except KeyError:
        pass

    details = ImportDetails()

    for node in module.body:  # pragma: no branch
        if (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Constant)
            and isinstance(node.value.value, str)
        ):
            # docstring
            continue

        if not isinstance(node, (ast.Import, ast.ImportFrom)):
            break

        if isinstance(node, ast.Import):
            for alias in node.names:
                if alias.name == "datetime":
                    if alias.asname is None:
                        details.datetime_module = "datetime"
                    else:
                        details.datetime_module = alias.asname

        # coverage bug
        # https://github.com/nedbat/coveragepy/issues/1333
        elif (  # pragma: no cover
            is_rewritable_import_from(node)
            and node.module == "django.utils.timezone"
            and any(a.name == "utc" for a in node.names)
        ):
            details.old_utc_import = node

    import_details[state] = details
    return details


def maybe_rewrite_import(
    details: ImportDetails,
) -> Iterable[tuple[Offset, TokenFunc]]:
    if details.rewrite_scheduled:
        return

    assert details.old_utc_import is not None
    yield (
        ast_start_offset(details.old_utc_import),
        partial(
            update_import_names,
            node=details.old_utc_import,
            name_map={"utc": ""},
        ),
    )

    details.rewrite_scheduled = True



"""
Replace imports from django.utils.translation:
https://docs.djangoproject.com/en/3.0/releases/3.0/#features-deprecated-in-3-0
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    extract_indent,
    find_and_replace_name,
    insert,
    update_import_names,
)

fixer = Fixer(
    __name__,
    min_version=(3, 0),
)

MODULE = "django.utils.text"
OLD_NAME = "unescape_entities"
NAME_MAP = {
    "unescape_entities": "",
}


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        node.level == 0
        and node.module == MODULE
        and any(
            (alias.name == OLD_NAME and alias.asname is None) for alias in node.names
        )
    ):
        yield ast_start_offset(node), partial(fix_import, node=node)


def fix_import(tokens: list[Token], i: int, *, node: ast.ImportFrom) -> None:
    j, indent = extract_indent(tokens, i)
    update_import_names(tokens, i, node=node, name_map={OLD_NAME: ""})
    insert(tokens, j, new_src=f"{indent}import html\n")


@fixer.register(ast.Name)
def visit_Name(
    state: State,
    node: ast.Name,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if node.id == OLD_NAME and OLD_NAME in state.from_imports[MODULE]:
        yield (
            ast_start_offset(node),
            partial(find_and_replace_name, name=OLD_NAME, new="html.escape"),
        )



from __future__ import annotations

import ast
import warnings
from typing import TYPE_CHECKING, Literal, cast

from tokenize_rt import Offset

if TYPE_CHECKING:
    from django_upgrade.data import State


def ast_parse(contents_text: str) -> ast.Module:
    # intentionally ignore warnings, we can't do anything about them
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return ast.parse(contents_text.encode())


def ast_start_offset(node: ast.expr | ast.keyword | ast.stmt) -> Offset:
    return Offset(node.lineno, node.col_offset)


def is_rewritable_import_from(node: ast.ImportFrom) -> bool:
    # Not relative import or import *
    return node.level == 0 and not (len(node.names) == 1 and node.names[0].name == "*")


TEST_CLIENT_REQUEST_METHODS = frozenset(
    (
        "request",
        "get",
        "post",
        "head",
        "options",
        "put",
        "patch",
        "delete",
        "trace",
    )
)


def looks_like_test_client_call(
    node: ast.AST, client_name: Literal["async_client", "client"]
) -> bool:
    return (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Attribute)
        and node.func.attr in TEST_CLIENT_REQUEST_METHODS
        and isinstance(node.func.value, ast.Attribute)
        and node.func.value.attr == client_name
        and isinstance(node.func.value.value, ast.Name)
        and node.func.value.value.id == "self"
    )


def is_passing_comparison(
    test: ast.Compare, state: State
) -> Literal["pass", "fail", None]:
    """
    Return whether the given ast.Compare node compares a version tuple with
    django.VERSION and would pass or fail for the current target version, or
    None if no match or cannot determine.
    """
    if not (
        isinstance(left := test.left, ast.Attribute)
        and isinstance(left.value, ast.Name)
        and left.value.id == "django"
        and left.attr == "VERSION"
        and len(test.ops) == 1
        and isinstance(test.ops[0], (ast.Gt, ast.GtE, ast.Lt, ast.LtE))
        and len(test.comparators) == 1
        and isinstance((comparator := test.comparators[0]), ast.Tuple)
        and 1 <= len(comparator.elts) <= 2
        and all(isinstance(e, ast.Constant) for e in comparator.elts)
        and all(isinstance(cast(ast.Constant, e).value, int) for e in comparator.elts)
    ):
        return None

    min_version = tuple(cast(ast.Constant, e).value for e in comparator.elts)
    if isinstance(test.ops[0], ast.Gt):
        if state.settings.target_version > min_version:
            return "pass"
    elif isinstance(test.ops[0], ast.GtE):
        if state.settings.target_version >= min_version:
            return "pass"
    elif isinstance(test.ops[0], ast.Lt):
        if state.settings.target_version >= min_version:
            return "fail"
    else:  # ast.LtE
        if state.settings.target_version > min_version:
            return "fail"
    return None



"""
Add the 'length' argument to get_random_string():
https://docs.djangoproject.com/en/3.1/releases/3.1/#features-deprecated-in-3-1
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import CODE, OP, find

fixer = Fixer(
    __name__,
    min_version=(3, 1),
)

MODULE = "django.utils.crypto"
NAME = "get_random_string"


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            (
                isinstance(node.func, ast.Name)
                and NAME in state.from_imports[MODULE]
                and node.func.id == NAME
            )
            or (
                isinstance(node.func, ast.Attribute)
                and node.func.attr == NAME
                and "crypto" in state.from_imports["django.utils"]
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "crypto"
            )
        )
        and len(node.args) == 0
        and not any(k.arg == "length" for k in node.keywords)
    ):
        yield (
            ast_start_offset(node),
            partial(
                add_length_argument,
                has_kwargs=(len(node.keywords) > 0),
            ),
        )


def add_length_argument(tokens: list[Token], i: int, *, has_kwargs: bool) -> None:
    j = find(tokens, i, name=OP, src="(")
    new_src = "length=12"
    if has_kwargs:
        new_src += ", "
    tokens.insert(j + 1, Token(name=CODE, src=new_src))



"""
Remove the DEFAULT_AUTO_FIELD setting if set to the new default of
"django.db.models.BigAutoField":
https://docs.djangoproject.com/en/6.0/releases/6.0/#default-auto-field-setting-now-defaults-to-bigautofield
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import erase_node

fixer = Fixer(
    __name__,
    min_version=(6, 0),
    condition=lambda state: state.looks_like_settings_file,
)


@fixer.register(ast.Assign)
def visit_Assign(
    state: State,
    node: ast.Assign,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        len(node.targets) == 1
        and isinstance(node.targets[0], ast.Name)
        and isinstance(node.value, ast.Constant)
        and isinstance(node.value.value, str)
        and node.value.value == "django.db.models.BigAutoField"
        and (
            state.looks_like_settings_file
            and node.targets[0].id == "DEFAULT_AUTO_FIELD"
            and (
                isinstance(parents[-1], ast.Module)
                or (
                    isinstance(parents[-1], ast.ClassDef)
                    and len(parents[-1].body) > 1
                    and isinstance(parents[-2], ast.Module)
                )
            )
        )
    ):
        yield ast_start_offset(node), partial(erase_node, node=node)



"""
Rewrite imports of StringAgg from django.contrib.postgres.aggregates to django.db.models,
wrapping string literal delimiter arguments in Value(), but only if all StringAgg calls
in the file have safe delimiter arguments:
https://docs.djangoproject.com/en/dev/releases/6.0/#models
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial
from weakref import WeakKeyDictionary

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import update_import_modules

fixer = Fixer(
    __name__,
    min_version=(6, 0),
)


do_rewrite: WeakKeyDictionary[ast.Module, bool] = WeakKeyDictionary()


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            node.module == "django.contrib.postgres.aggregates"
            or node.module == "django.contrib.postgres.aggregates.general"
        )
        and is_rewritable_import_from(node)
        and any(
            alias.name == "StringAgg" and alias.asname is None for alias in node.names
        )
    ):
        module = parents[0]
        assert isinstance(module, ast.Module)
        do_rewrite.setdefault(module, True)

        yield (
            ast_start_offset(node),
            partial(rewrite_import_from, node=node, module=module),
        )


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(node.func, ast.Name)
        and node.func.id == "StringAgg"
        and (
            node.func.id in state.from_imports["django.contrib.postgres.aggregates"]
            or node.func.id
            in state.from_imports["django.contrib.postgres.aggregates.general"]
        )
    ):
        module = parents[0]
        assert isinstance(module, ast.Module)

        if do_rewrite.get(module) is not True:
            return

        delimiter = None
        if len(node.args) == 1:
            for kw in node.keywords:
                if kw.arg == "delimiter":
                    delimiter = kw.value
                    break
        elif len(node.args) >= 2:
            delimiter = node.args[1]

        if delimiter is None:
            # Cannot rewrite
            do_rewrite[module] = False
            return

        if isinstance(delimiter, ast.Constant) and isinstance(delimiter.value, str):
            if "Value" in state.from_imports["django.db.models"]:
                wrap = "Value"
            elif "models" in state.from_imports["django.db"]:
                wrap = "models.Value"
            else:
                do_rewrite[module] = False
                return

            yield (
                ast_start_offset(delimiter),
                partial(
                    wrap_delimiter,
                    wrap=wrap,
                ),
            )
        elif (
            isinstance(delimiter, ast.Call)
            and isinstance(delimiter.func, ast.Name)
            and delimiter.func.id == "Value"
            and "Value" in state.from_imports["django.db.models"]
        ):
            # All good.
            pass
        else:
            do_rewrite[module] = False


def rewrite_import_from(
    tokens: list[Token], i: int, *, node: ast.ImportFrom, module: ast.Module
) -> None:
    if do_rewrite.get(module) is not True:
        return

    update_import_modules(
        tokens, i, node=node, module_rewrites={"StringAgg": "django.db.models"}
    )


def wrap_delimiter(tokens: list[Token], i: int, *, wrap: str) -> None:
    tokens[i] = tokens[i]._replace(src=f"{wrap}(" + tokens[i].src + ")")



"""
Update URL definitions:
https://docs.djangoproject.com/en/2.0/releases/2.0/#simplified-url-routing-syntax
"""

from __future__ import annotations

import ast
import re
from collections.abc import Iterable, MutableMapping
from functools import partial
from weakref import WeakKeyDictionary

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    STRING,
    extract_indent,
    find,
    find_last_token,
    insert,
    replace,
    str_repr_matching,
    update_import_names,
)

fixer = Fixer(
    __name__,
    min_version=(2, 0),
)


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        node.module == "django.conf.urls"
        and is_rewritable_import_from(node)
        and any(alias.name in ("include", "url") for alias in node.names)
    ):
        yield (
            ast_start_offset(node),
            partial(
                update_django_conf_import,
                node=node,
                state=state,
            ),
        )
    elif (
        node.module == "django.urls"
        and is_rewritable_import_from(node)
        and any(alias.name == "re_path" for alias in node.names)
    ):
        yield (
            ast_start_offset(node),
            partial(
                update_django_urls_import,
                node=node,
                state=state,
            ),
        )


# Track if re_path has been used, and which names need adding.
# When fixing import statements, these variables determine which names to
# import/remove.
state_re_path_used: MutableMapping[State, bool] = WeakKeyDictionary()
state_added_names: MutableMapping[State, set[str]] = WeakKeyDictionary()


def update_django_conf_import(
    tokens: list[Token], i: int, *, node: ast.ImportFrom, state: State
) -> None:
    re_path_imported = "re_path" in state.from_imports["django.urls"]
    added_names = state_added_names.pop(state, set())
    removals = set()

    for alias in node.names:
        if alias.asname is not None:
            continue
        if alias.name in ("include", "url") and (added_names or re_path_imported):
            removals.add(alias.name)

    if removals:
        j, indent = extract_indent(tokens, i)
        update_import_names(
            tokens,
            i,
            node=node,
            name_map=dict.fromkeys(removals, ""),
        )
        if not re_path_imported:
            joined_names = ", ".join(sorted(added_names))
            insert(
                tokens,
                j,
                new_src=f"{indent}from django.urls import {joined_names}\n",
            )
        else:
            state_added_names[state] = added_names


def update_django_urls_import(
    tokens: list[Token], i: int, *, node: ast.ImportFrom, state: State
) -> None:
    re_path_used = state_re_path_used.get(state, False)
    added_names = state_added_names.pop(state, set())
    missing_names = added_names - state.from_imports["django.urls"]

    if (
        added_names
        and not re_path_used
        and "re_path" in state.from_imports["django.urls"]
    ):
        update_import_names(
            tokens,
            i,
            node=node,
            name_map={"re_path": ""},
        )

    if missing_names:
        j, indent = extract_indent(tokens, i)
        joined_names = ", ".join(sorted(missing_names))
        insert(
            tokens,
            j,
            new_src=f"{indent}from django.urls import {joined_names}\n",
        )


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if isinstance(node.func, ast.Name):
        if (
            (
                (
                    (node_name := node.func.id) == "url"
                    and "url" in state.from_imports["django.conf.urls"]
                )
                or (
                    (node_name := node.func.id) == "re_path"
                    and "re_path" in state.from_imports["django.urls"]
                )
            )
            # cannot convert where called with all kwargs as names don't align
            and len(node.args) >= 1
        ):
            regex_path: ast.Constant | None = None
            if isinstance(node.args[0], ast.Constant) and isinstance(
                node.args[0].value, str
            ):
                regex_path = node.args[0]

            include_called = (
                len(node.args) >= 2
                and isinstance(node.args[1], ast.Call)
                and isinstance(node.args[1].func, ast.Name)
                and node.args[1].func.id == "include"
            )
            yield (
                ast_start_offset(node),
                partial(
                    fix_url_call,
                    regex_path=regex_path,
                    state=state,
                    node_name=node_name,
                    include_called=include_called,
                ),
            )

        elif (
            node.func.id == "include"
            and "include" in state.from_imports["django.conf.urls"]
        ):
            state_added_names.setdefault(state, set()).add("include")


def fix_url_call(
    tokens: list[Token],
    i: int,
    *,
    regex_path: ast.Constant | None,
    state: State,
    node_name: str,
    include_called: bool,
) -> None:
    new_name = "re_path"
    if regex_path is not None:
        regex_value = regex_path.value
        assert isinstance(regex_value, str)
        path = convert_path_syntax(regex_value, include_called)
        if path is not None:
            string_start_idx = find(tokens, i, name=STRING)
            string_end_idx = find_last_token(tokens, string_start_idx, node=regex_path)
            path = str_repr_matching(path, match_quotes=tokens[string_start_idx].src)
            tokens[string_start_idx : string_end_idx + 1] = [Token(STRING, path)]
            new_name = "path"
    if new_name != node_name:
        state_added_names.setdefault(state, set()).add(new_name)
        replace(tokens, i, src=new_name)
    else:
        state_re_path_used.setdefault(state, True)


REGEX_TO_CONVERTER = {
    "[0-9]+": "int",
    r"\d+": "int",
    ".+": "path",
    "[-a-zA-Z0-9_]+": "slug",
    "[^/]+": "str",
    "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}": "uuid",
}


def convert_path_syntax(regex_path: str, include_called: bool) -> str | None:
    if not (regex_path.endswith("$") or include_called):
        return None
    remaining = regex_path.removeprefix("^")
    remaining = remaining.removesuffix("$")
    path = ""
    while "(?P<" in remaining:
        prefix, rest = remaining.split("(?P<", 1)
        group, remaining = rest.split(")", 1)
        group_name, group_regex = group.split(">", 1)
        if not group_name.isidentifier():
            return None
        try:
            converter = REGEX_TO_CONVERTER[group_regex]
        except KeyError:
            return None

        path += prefix
        path += f"<{converter}:{group_name}>"

    path += remaining

    path = re.sub(r"\\\.", ".", path)  # unescape literal dots

    if not re.fullmatch(r"[a-zA-Z0-9_\-./<>:]*", path):
        # path still contains regex special characters
        return None

    return path



"""
Rewrite django.db.models.NullBooleanField to BooleanField:
https://docs.djangoproject.com/en/3.1/releases/3.1/#features-deprecated-in-3-1
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    CODE,
    OP,
    find,
    find_and_replace_name,
    parse_call_args,
    update_import_names,
)

fixer = Fixer(
    __name__,
    min_version=(3, 1),
    condition=lambda state: state.looks_like_models_file,
)


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if is_rewritable_import_from(node) and node.module == "django.db.models":
        yield (
            ast_start_offset(node),
            partial(
                update_import_names,
                node=node,
                name_map={"NullBooleanField": "BooleanField"},
            ),
        )


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(node.func, ast.Name)
        and "NullBooleanField" in state.from_imports["django.db.models"]
        and node.func.id == "NullBooleanField"
    ) or (
        isinstance(node.func, ast.Attribute)
        and node.func.attr == "NullBooleanField"
        and "models" in state.from_imports["django.db"]
        and isinstance(node.func.value, ast.Name)
        and node.func.value.id == "models"
    ):
        yield ast_start_offset(node), partial(fix_null_boolean_field, node=node)


def fix_null_boolean_field(tokens: list[Token], i: int, *, node: ast.Call) -> None:
    if not any(k.arg == "null" for k in node.keywords):
        j = find(tokens, i, name=OP, src="(")
        func_args, j = parse_call_args(tokens, j)

        new_src = "null=True"
        if len(func_args) > 0:
            new_src = " " + new_src
            final_start_idx, final_end_idx = func_args[-1]
            final_has_comma = any(
                t.name == OP and t.src == ","
                for t in tokens[final_start_idx : final_end_idx + 1]
            )
            if not final_has_comma:
                new_src = "," + new_src

        tokens.insert(j - 1, Token(name=CODE, src=new_src))

    find_and_replace_name(tokens, i, name="NullBooleanField", new="BooleanField")



from __future__ import annotations

import argparse
import re
import sys
import tokenize
from collections.abc import Sequence
from importlib import metadata
from typing import Any, cast

from tokenize_rt import (
    UNIMPORTANT_WS,
    Token,
    reversed_enumerate,
    src_to_tokens,
    tokens_to_src,
)

from django_upgrade.ast import ast_parse
from django_upgrade.data import FIXERS, Settings, visit
from django_upgrade.tokens import DEDENT

SUPPORTED_TARGET_VERSIONS = {
    (1, 7),
    (1, 8),
    (1, 9),
    (1, 10),
    (1, 11),
    (2, 0),
    (2, 1),
    (2, 2),
    (3, 0),
    (3, 1),
    (3, 2),
    (4, 0),
    (4, 1),
    (4, 2),
    (5, 0),
    (5, 1),
    (5, 2),
    (6, 0),
}


def main(argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser(prog="django-upgrade")
    parser.suggest_on_error = True  # type: ignore[attr-defined]
    parser.add_argument(
        "filenames", nargs="+", help="Filenames to fix, or '-' for stdin."
    )
    parser.add_argument(
        "--target-version",
        default="auto",
        choices=[
            "auto",
            *[f"{major}.{minor}" for major, minor in SUPPORTED_TARGET_VERSIONS],
        ],
        help="The version of Django to target.",
    )
    parser.add_argument(
        "--check",
        action="store_true",
        help="Only output files to change, do not change files.",
    )
    parser.add_argument(
        "--exit-zero-even-if-changed",
        action="store_true",
        help="Exit with a zero return code even if files have changed.",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=metadata.version("django-upgrade"),
        help="Show the version number and exit.",
    )
    parser.add_argument(
        "--only",
        action="append",
        type=fixer_type,
        help="Run only the selected fixers.",
    )
    parser.add_argument(
        "--skip",
        action="append",
        type=fixer_type,
        help="Skip the selected fixers.",
    )
    parser.add_argument(
        "--list-fixers", nargs=0, action=ListFixersAction, help="List all fixer names."
    )

    args = parser.parse_args(argv)

    settings = Settings(
        target_version=get_target_version(args.target_version),
        only_fixers=set(args.only) if args.only else None,
        skip_fixers=set(args.skip) if args.skip else None,
    )

    ret = 0
    for filename in args.filenames:
        ret |= fix_file(
            filename,
            settings,
            exit_zero_even_if_changed=args.exit_zero_even_if_changed,
            check=args.check,
        )

    return ret


def fixer_type(string: str) -> str:
    if string not in FIXERS:
        raise argparse.ArgumentTypeError(f"Unknown fixer: {string!r}")
    return string


class ListFixersAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        for name in sorted(FIXERS):
            print(name)
        parser.exit()


def get_target_version(string: str) -> tuple[int, int]:
    default = (2, 2)
    if string != "auto":
        return cast(
            tuple[int, int],
            tuple(int(x) for x in string.split(".", 1)),
        )

    if sys.version_info < (3, 11):
        return default

    import tomllib

    try:
        with open("pyproject.toml", "rb") as fp:
            config = tomllib.load(fp)
    except FileNotFoundError:
        return default

    deps = config.get("project", {}).get("dependencies", [])
    for dep in deps:
        match = re.fullmatch(
            r"""
            django
            \s*
            (
                \[[^]]+\]
                \s*
            )?
            (?:==|~=|>=)
            \s*
            (
                (?P<major>[0-9]+)
                \.
                (?P<minor>[0-9]+)
                (
                    (?:a|b|rc)
                    [0-9]+
                |
                    \.
                    [0-9]+
                |
                )
            )
            (
                \s*,\s*
                (<|<=)
                \s*
                [0-9]+
                (
                    \.
                    [0-9]+
                    (
                        \.
                        [0-9]+
                    )?
                )?
            )?
            """,
            dep.lower(),
            re.VERBOSE,
        )
        if match:
            major = int(match["major"])
            minor = int(match["minor"])
            if (major, minor) in SUPPORTED_TARGET_VERSIONS:
                print(
                    f"Detected Django version from pyproject.toml: {major}.{minor}",
                    file=sys.stderr,
                )
                return (major, minor)

    return default


def fix_file(
    filename: str,
    settings: Settings,
    exit_zero_even_if_changed: bool,
    check: bool,
) -> int:
    if filename == "-":
        contents_bytes = sys.stdin.buffer.read()
    else:
        with open(filename, "rb") as fb:
            contents_bytes = fb.read()

    try:
        contents_text_orig = contents_text = contents_bytes.decode()
    except UnicodeDecodeError:
        print(f"{filename} is non-utf-8 (not supported)", file=sys.stderr)
        return 1

    contents_text = apply_fixers(contents_text, settings, filename)

    returncode = 0
    if contents_text != contents_text_orig:
        if check:
            display_name = "stdin" if filename == "-" else filename
            print(f"Would rewrite {display_name}", file=sys.stderr)
            returncode = 1
        else:
            if filename == "-":
                print(contents_text, end="")
            else:
                print(f"Rewriting {filename}", file=sys.stderr)
                with open(filename, "w", encoding="UTF-8", newline="") as f:
                    f.write(contents_text)
                if not exit_zero_even_if_changed:
                    returncode = 1
    else:
        if filename == "-" and not check:
            print(contents_text, end="")

    return returncode


def apply_fixers(contents_text: str, settings: Settings, filename: str) -> str:
    try:
        ast_obj = ast_parse(contents_text)
    except SyntaxError:
        return contents_text

    callbacks = visit(ast_obj, settings, filename)

    if not callbacks:
        return contents_text

    try:
        tokens = src_to_tokens(contents_text)
    except tokenize.TokenError:  # pragma: no cover (bpo-2180)
        return contents_text

    fixup_dedent_tokens(tokens)

    for i, token in reversed_enumerate(tokens):
        if not token.src:
            continue
        # though this is a defaultdict, by using `.get()` this function's
        # self time is almost 50% faster
        for callback in callbacks.get(token.offset, ()):
            callback(tokens, i)

    # no types for tokenize-rt
    return tokens_to_src(tokens)  # type: ignore [no-any-return]


def fixup_dedent_tokens(tokens: list[Token]) -> None:
    """For whatever reason the DEDENT / UNIMPORTANT_WS tokens are misordered

    | if True:
    |     if True:
    |         pass
    |     else:
    |^    ^- DEDENT
    |+----UNIMPORTANT_WS
    """
    for i, token in enumerate(tokens):
        if token.name == UNIMPORTANT_WS and tokens[i + 1].name == DEDENT:
            tokens[i], tokens[i + 1] = tokens[i + 1], tokens[i]



"""
Rewrite Model.Meta.index_together declarations into Model.Meta.Index
declarations.
https://docs.djangoproject.com/en/4.2/releases/4.2/#index-together-option-is-deprecated-in-favor-of-indexes
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import UNIMPORTANT_WS, Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    DEDENT,
    INDENT,
    OP,
    PHYSICAL_NEWLINE,
    erase_node,
    extract_indent,
    find_last_token,
    insert,
    str_repr_matching,
)

fixer = Fixer(
    __name__,
    min_version=(4, 2),
    condition=lambda state: state.looks_like_models_file,
)


@fixer.register(ast.ClassDef)
def visit_ClassDef(
    state: State,
    node: ast.ClassDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        node.name != "Meta"
        or sum(isinstance(p, ast.ClassDef) for p in parents[1:]) != 1
        or not all(isinstance(subnode, ast.Assign) for subnode in node.body)
    ):
        return

    # Find rewritable index_together declaration
    index_togethers: list[ast.Assign] = []
    for subnode in node.body:
        assert isinstance(subnode, ast.Assign)  # type checked above
        if (
            len(subnode.targets) == 1
            and isinstance(subnode.targets[0], ast.Name)
            and subnode.targets[0].id == "index_together"
            and isinstance(subnode.value, (ast.List, ast.Tuple))
            and (
                all(
                    (isinstance(elt, ast.Constant) and isinstance(elt.value, str))
                    for elt in subnode.value.elts
                )
                or all(
                    isinstance(elt, (ast.List, ast.Tuple))
                    and all(
                        (
                            isinstance(subelt, ast.Constant)
                            and isinstance(subelt.value, str)
                        )
                        for subelt in elt.elts
                    )
                    for elt in subnode.value.elts
                )
            )
        ):
            index_togethers.append(subnode)

    if len(index_togethers) != 1:
        return

    index_together = index_togethers[0]

    # Try to find an indexes declaration to extend
    indexeses: list[ast.Assign] = []
    for subnode in node.body:
        assert isinstance(subnode, ast.Assign)  # type checked above
        if (
            len(subnode.targets) == 1
            and isinstance(subnode.targets[0], ast.Name)
            and subnode.targets[0].id == "indexes"
            and isinstance(subnode.value, (ast.List, ast.Tuple))
        ):
            indexeses.append(subnode)

    if len(indexeses) > 1:
        return

    try:
        indexes = indexeses[0]
    except IndexError:
        indexes = None

    if (
        "models" in state.from_imports["django.db"]
        or "models" in state.from_imports["django.contrib.gis.db"]
    ):
        index_ref = "models.Index"
    elif (
        "Index" in state.from_imports["django.db.models"]
        or "Index" in state.from_imports["django.contrib.gis.db.models"]
    ):
        index_ref = "Index"
    else:
        return

    src_chunks = []
    assert isinstance(index_together.value, (ast.List, ast.Tuple))  # type checked above
    # Single index: pretend it was a list-of-lists (or tuple-of-tuples, etc.)
    if isinstance(index_together.value.elts[0], ast.Constant):
        iterate: list[ast.List | ast.Tuple] = [
            type(index_together.value)(elts=index_together.value.elts)
        ]
    else:
        # type checked above
        iterate = index_together.value.elts  # type: ignore [assignment]
    for indexnode in iterate:
        index_src = index_ref
        index_src += "(fields="
        if isinstance(indexnode, ast.Tuple):
            index_src += "("
        else:
            index_src += "["

        assert isinstance(indexnode, (ast.List, ast.Tuple))  # type checked above
        for const in indexnode.elts:
            # type checked above:
            assert isinstance(const, ast.Constant)
            assert isinstance(const.value, str)
            # Default to double quotes because they’re fashionable
            index_src += str_repr_matching(const.value, match_quotes='"')
            index_src += ", "

        index_src = index_src.removesuffix(", ")

        if isinstance(indexnode, ast.Tuple):
            index_src += ")"
        else:
            index_src += "]"

        index_src += ")"

        src_chunks.append(index_src)

    index_src = ", ".join(src_chunks)

    yield (
        ast_start_offset(index_together),
        partial(
            remove_index_together_and_maybe_add_indexes,
            index_together=index_together,
            add_indexes=(indexes is None),
            index_src=index_src,
        ),
    )
    if indexes is not None:
        yield (
            ast_start_offset(indexes),
            partial(extend_indexes, indexes=indexes, index_src=index_src),
        )


def remove_index_together_and_maybe_add_indexes(
    tokens: list[Token],
    i: int,
    *,
    index_together: ast.Assign,
    add_indexes: bool,
    index_src: str,
) -> None:
    j, indent = extract_indent(tokens, i)
    erase_node(tokens, i, node=index_together)
    if add_indexes:
        insert(tokens, j, new_src=f"{indent}indexes = [{index_src}]\n")


def extend_indexes(
    tokens: list[Token],
    i: int,
    *,
    indexes: ast.Assign,
    index_src: str,
) -> None:
    assert isinstance(indexes.value, (ast.List, ast.Tuple))  # type checked above
    closing_punctuation = find_last_token(tokens, i, node=indexes.value)
    if len(indexes.value.elts) == 0 or tokens[closing_punctuation - 1].name in (
        INDENT,
        DEDENT,
        UNIMPORTANT_WS,
        PHYSICAL_NEWLINE,
    ):
        prefix = ""
    else:
        j = find_last_token(tokens, i, node=indexes.value.elts[-1])
        if any(
            t.name == OP and t.src == "," for t in tokens[j + 1 : closing_punctuation]
        ):
            prefix = " "
        else:
            prefix = ", "

    insert(tokens, closing_punctuation, new_src=f"{prefix}{index_src}")



"""
Replace imports from django.utils.encoding:
https://docs.djangoproject.com/en/3.0/releases/3.0/#django-utils-encoding-force-text-and-smart-text  # noqa: E501
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset, is_rewritable_import_from
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import find_and_replace_name, update_import_names

fixer = Fixer(
    __name__,
    min_version=(3, 0),
)

MODULE = "django.utils.encoding"
NAMES = {
    "force_text": "force_str",
    "smart_text": "smart_str",
}


@fixer.register(ast.ImportFrom)
def visit_ImportFrom(
    state: State,
    node: ast.ImportFrom,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if node.module == MODULE and is_rewritable_import_from(node):
        yield (
            ast_start_offset(node),
            partial(update_import_names, node=node, name_map=NAMES),
        )


@fixer.register(ast.Name)
def visit_Name(
    state: State,
    node: ast.Name,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (name := node.id) in NAMES and name in state.from_imports[MODULE]:
        yield (
            ast_start_offset(node),
            partial(find_and_replace_name, name=name, new=NAMES[name]),
        )


@fixer.register(ast.Attribute)
def visit_Attribute(
    state: State,
    node: ast.Attribute,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (name := node.attr) in NAMES
        and isinstance(node.value, ast.Name)
        and node.value.id == "encoding"
        and "encoding" in state.from_imports["django.utils"]
    ):
        yield (
            ast_start_offset(node),
            partial(find_and_replace_name, name=name, new=NAMES[name]),
        )



"""
Drop `.choices` for model field `choices` parameters:
https://docs.djangoproject.com/en/5.0/releases/5.0/#forms
"""

from __future__ import annotations

import ast
from collections import defaultdict
from collections.abc import Iterable
from functools import partial
from typing import cast
from weakref import WeakKeyDictionary

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import OP, find_last_token, reverse_find

fixer = Fixer(
    __name__,
    min_version=(5, 0),
    condition=lambda state: state.looks_like_models_file,
)

# Cache defined enumeration types by module
module_defined_enumeration_types: WeakKeyDictionary[ast.Module, dict[str, int]] = (
    WeakKeyDictionary()
)


def defined_enumeration_types(module: ast.Module, up_to_line: int) -> set[str]:
    """
    Return a set of enumeration type class names defined in the given module, up to a line number.
    """
    if module not in module_defined_enumeration_types:
        enum_dict = {}
        from_imports: defaultdict[str, set[str]] = defaultdict(set)
        for node in module.body:
            if (
                isinstance(node, ast.ImportFrom)
                and node.level == 0
                and node.module is not None
            ):
                from_imports[node.module].update(
                    name.name
                    for name in node.names
                    if name.asname is None and name.name != "*"
                )
            elif isinstance(node, ast.ClassDef):
                # Check if the class inherits from one of Django's choice types
                for base in node.bases:
                    if _is_django_choices_type(from_imports, base):
                        enum_dict[node.name] = node.lineno
                        break
        module_defined_enumeration_types[module] = enum_dict

    return {
        name
        for name, line in module_defined_enumeration_types[module].items()
        if line <= up_to_line
    }


DJANGO_CHOICES_TYPES = {
    "TextChoices",
    "IntegerChoices",
    "Choices",
}


def _is_django_choices_type(
    from_imports: defaultdict[str, set[str]], node: ast.expr
) -> bool:
    """Check if an AST node refers to a Django enumeration type base class."""
    return (
        isinstance(node, ast.Name)
        and node.id in DJANGO_CHOICES_TYPES
        and (
            node.id in from_imports["django.db.models"]
            or node.id in from_imports["django.db.models.enums"]
        )
    ) or (
        isinstance(node, ast.Attribute)
        and node.attr in DJANGO_CHOICES_TYPES
        and isinstance(node.value, ast.Name)
        and (
            (node.value.id == "models" and node.value.id in from_imports["django.db"])
            or (
                node.value.id == "enums"
                and node.value.id in from_imports["django.db.models"]
            )
        )
    )


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.attr.endswith("Field")
        )
        or (isinstance(node.func, ast.Name) and node.func.id.endswith("Field"))
    ) and any(
        kw.arg == "choices"
        and isinstance(kw.value, ast.Attribute)
        and (target_node := kw.value).attr == "choices"
        and isinstance(target_node.value, ast.Name)
        and (
            target_node.value.id
            in defined_enumeration_types(
                cast(ast.Module, parents[0]),
                node.lineno,
            )
        )
        for kw in node.keywords
    ):
        yield ast_start_offset(target_node), partial(remove_choices, node=target_node)


def remove_choices(tokens: list[Token], i: int, node: ast.Attribute) -> None:
    j = find_last_token(tokens, i, node=node)
    i = reverse_find(tokens, j, name=OP, src=".")
    del tokens[i : j + 1]



"""
Rewrite the ADMINS and MANAGERS settings to change lists of tuples to lists of strings:
https://docs.djangoproject.com/en/6.0/releases/6.0/#positional-arguments-in-django-core-mail-apis:~:text=Setting%20ADMINS%20or%20MANAGERS
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import find_first_token, find_last_token

fixer = Fixer(
    __name__,
    min_version=(6, 0),
    condition=lambda state: state.looks_like_settings_file,
)


@fixer.register(ast.Assign)
def visit_Assign(
    state: State,
    node: ast.Assign,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        len(node.targets) == 1
        and isinstance(node.targets[0], ast.Name)
        and node.targets[0].id in ("ADMINS", "MANAGERS")
        and isinstance(node.value, (ast.List, ast.Tuple))
        and all(
            isinstance(elt, ast.Tuple) and len(elt.elts) == 2 for elt in node.value.elts
        )
        and (
            isinstance(parents[-1], ast.Module)
            or (
                isinstance(parents[-1], ast.ClassDef)
                and len(parents[-1].body) > 1
                and isinstance(parents[-2], ast.Module)
            )
        )
    ):
        yield ast_start_offset(node), partial(update_setting, node=node)


def update_setting(
    tokens: list[Token],
    i: int,
    *,
    node: ast.Assign,
) -> None:
    assert isinstance(node.value, (ast.List, ast.Tuple))

    indexes = []
    j = i
    for elt in node.value.elts:
        tuple_start = find_first_token(tokens, j, node=elt)
        tuple_end = find_last_token(tokens, tuple_start, node=elt)

        assert isinstance(elt, ast.Tuple) and len(elt.elts) == 2
        address = elt.elts[1]
        address_start = find_first_token(tokens, tuple_start, node=address)
        address_end = find_last_token(tokens, address_start, node=address)

        indexes.append((tuple_start, tuple_end, address_start, address_end))

        j = tuple_end + 1

    for tuple_start, tuple_end, address_start, address_end in reversed(indexes):
        # Replace the entire tuple with just the email address
        tokens[tuple_start : tuple_end + 1] = tokens[address_start : address_end + 1]



"""
Remove the 'providing_args' argument from Signal():
https://docs.djangoproject.com/en/3.1/releases/3.1/#features-deprecated-in-3-1
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import UNIMPORTANT_WS, Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    CODE,
    COMMENT,
    INDENT,
    OP,
    consume,
    find,
    parse_call_args,
    reverse_consume,
)

fixer = Fixer(
    __name__,
    min_version=(3, 1),
)

MODULE = "django.dispatch"
NAME = "Signal"


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        (
            isinstance(node.func, ast.Name)
            and NAME in state.from_imports[MODULE]
            and node.func.id == NAME
        )
        or (
            isinstance(node.func, ast.Attribute)
            and node.func.attr == NAME
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "dispatch"
            and "dispatch" in state.from_imports["django"]
        )
    ) and (len(node.args) > 0 or any(k.arg == "providing_args" for k in node.keywords)):
        yield (
            ast_start_offset(node),
            partial(
                remove_providing_args,
                node=node,
            ),
        )


def remove_providing_args(tokens: list[Token], i: int, *, node: ast.Call) -> None:
    j = find(tokens, i, name=OP, src="(")
    func_args, _ = parse_call_args(tokens, j)

    if len(node.args):
        start_idx, end_idx = func_args[0]
        if len(node.args) == 1:
            del tokens[start_idx:end_idx]
        else:
            # Have to replace with None
            tokens[start_idx:end_idx] = [Token(name=CODE, src="None")]
    else:
        for n, keyword in enumerate(node.keywords):
            if keyword.arg == "providing_args":
                start_idx, end_idx = func_args[n]

                start_idx = reverse_consume(tokens, start_idx, name=UNIMPORTANT_WS)
                start_idx = reverse_consume(tokens, start_idx, name=INDENT)
                if n > 0:
                    start_idx = reverse_consume(tokens, start_idx, name=OP, src=",")

                if n < len(node.keywords) - 1:
                    end_idx = consume(tokens, end_idx, name=UNIMPORTANT_WS)
                    end_idx = consume(tokens, end_idx, name=OP, src=",")
                    end_idx = consume(tokens, end_idx, name=UNIMPORTANT_WS)
                    end_idx = consume(tokens, end_idx, name=COMMENT)
                    end_idx += 1

                del tokens[start_idx:end_idx]



"""
PASSWORD_RESET_TIMEOUT_DAYS setting replaced with PASSWORD_RESET_TIMEOUT:
https://docs.djangoproject.com/en/3.1/releases/3.1/#django-contrib-auth
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import CODE, OP, find

fixer = Fixer(
    __name__,
    min_version=(3, 1),
    condition=lambda state: state.looks_like_settings_file,
)

OLD_NAME = "PASSWORD_RESET_TIMEOUT_DAYS"
NEW_NAME = "PASSWORD_RESET_TIMEOUT"


@fixer.register(ast.Assign)
def visit_Assign(
    state: State,
    node: ast.Assign,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        len(node.targets) == 1
        and isinstance(node.targets[0], ast.Name)
        and node.targets[0].id == OLD_NAME
    ):
        yield ast_start_offset(node), partial(rewrite_setting, node=node)


def rewrite_setting(tokens: list[Token], i: int, *, node: ast.Assign) -> None:
    tokens[i] = tokens[i]._replace(name=CODE, src=NEW_NAME)
    j = find(tokens, i, name=OP, src="=")
    tokens.insert(j + 1, Token(name=CODE, src=" 60 * 60 * 24 *"))



"""
Update settings.DATABASES backend path 'django.db.backends.postgresql_psycopg2'
to the new 'django.db.backends.postgresql'.
https://docs.djangoproject.com/en/2.0/releases/2.0/#id1
"""

from __future__ import annotations

import ast
from collections.abc import Iterable

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import replace, str_repr_matching

fixer = Fixer(
    __name__,
    min_version=(1, 9),
    condition=lambda state: state.looks_like_settings_file,
)


@fixer.register(ast.Dict)
def visit_Dict(
    state: State,
    node: ast.Dict,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        len(parents) >= 2
        and isinstance(parents[-1], ast.Dict)
        and isinstance((db_setting := parents[-2]), ast.Assign)
        and len(db_setting.targets) == 1
        and isinstance(db_setting.targets[0], ast.Name)
        and db_setting.targets[0].id == "DATABASES"
        and any(
            (
                isinstance(key, ast.Constant)
                and key.value == "ENGINE"
                and isinstance(val, ast.Constant)
                and (target_node := val).value
                == "django.db.backends.postgresql_psycopg2"
            )
            for key, val in zip(node.keys, node.values)
        )
    ):
        yield ast_start_offset(target_node), replace_engine


def replace_engine(tokens: list[Token], i: int) -> None:
    src = str_repr_matching("django.db.backends.postgresql", match_quotes=tokens[i].src)
    replace(tokens, i, src=src)



"""
Rewrite spelling of assertFormsetError and assertQuerysetEqual to include
capitalized “Set”:
https://docs.djangoproject.com/en/4.2/releases/4.2/#miscellaneous
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial

from tokenize_rt import Offset

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import find_and_replace_name

fixer = Fixer(
    __name__,
    min_version=(4, 2),
    condition=lambda state: state.looks_like_test_file,
)

MODULE = "django.test.testcase"
NAMES = {
    "assertFormsetError": "assertFormSetError",
    "assertQuerysetEqual": "assertQuerySetEqual",
}


@fixer.register(ast.Call)
def visit_Call(
    state: State,
    node: ast.Call,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(func := node.func, ast.Attribute)
        and (name := func.attr) in NAMES
        and isinstance(func.value, ast.Name)
        and func.value.id == "self"
    ):
        yield (
            ast_start_offset(func),
            partial(find_and_replace_name, name=name, new=NAMES[name]),
        )



"""
Update functions with admin attributes attached to use the new decorator forms:

https://docs.djangoproject.com/en/3.2/ref/contrib/admin/actions/#django.contrib.admin.action
https://docs.djangoproject.com/en/3.2/ref/contrib/admin/#django.contrib.admin.display
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial
from typing import Literal

from tokenize_rt import Offset, Token, tokens_to_src

from django_upgrade.ast import ast_start_offset
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import (
    OP,
    erase_node,
    extract_indent,
    find_last_token,
    insert,
    reverse_find,
)

fixer = Fixer(
    __name__,
    min_version=(3, 2),
)


@fixer.register(ast.Module)
def visit_Module(
    state: State,
    node: ast.Module,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    yield from visit_Module_or_ClassDef(state, node, parents)


@fixer.register(ast.ClassDef)
def visit_ClassDef(
    state: State,
    node: ast.ClassDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    yield from visit_Module_or_ClassDef(state, node, parents)


# Map from old assigned names to new decorator names, which were changed to be
# shorter
NAME_MAPS = {
    "action": {
        "short_description": "description",
        "allowed_permissions": "permissions",
    },
    "display": {
        "short_description": "description",
        "boolean": "boolean",
        "empty_value_display": "empty_value",
        "admin_order_field": "ordering",
    },
}


class FunctionDetails:
    __slots__ = ("node", "decorator", "assignments", "values")

    def __init__(self, node: ast.FunctionDef, decorator: Literal["action", "display"]):
        self.node = node
        self.decorator = decorator
        # Discovered attribute assignments to the given function
        self.assignments: dict[str, ast.Assign] = {}
        # Source strings for the values assigned to attributes, used in
        # adding the decorator
        self.values: dict[str, str] = {}


def visit_Module_or_ClassDef(
    state: State,
    node: ast.Module | ast.ClassDef,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    # Potential action and display functions to details of assigned attributes
    funcs: dict[str, FunctionDetails] = {}

    # Display functions take one arg (the model instance), but will also take
    # self inside classes
    if isinstance(node, ast.Module):
        display_func_args = 1
    else:
        display_func_args = 2

    # Check for 'from django.contrib import admin' from state.from_imports,
    # but also directly when visiting a module. state.from_imports isn’t
    # populated yet when visiting a module... (could fix by doing two passes?)
    admin_imported = (
        "admin" in state.from_imports["django.contrib"]
        or "admin" in state.from_imports["django.contrib.gis"]
    )

    for subnode in ast.iter_child_nodes(node):
        # coverage bug
        # https://github.com/nedbat/coveragepy/issues/1333
        if (  # pragma: no cover
            not admin_imported
            and isinstance(subnode, ast.ImportFrom)
            and subnode.module in ("django.contrib", "django.contrib.gis")
            and any(
                alias.name == "admin" and alias.asname is None
                for alias in subnode.names
            )
        ):
            admin_imported = True
        elif isinstance(subnode, ast.FunctionDef):
            if (
                # Django calls action functions with exactly three arguments,
                # positionally (modeladmin, request, queryset)
                (len(subnode.args.posonlyargs) + len(subnode.args.args)) == 3
                and len(subnode.args.kwonlyargs) == 0
                # TODO: check that no admin.action decorator already applied
            ):
                funcs[subnode.name] = FunctionDetails(subnode, "action")
            elif (
                (len(subnode.args.posonlyargs) + len(subnode.args.args))
                == display_func_args
                and len(subnode.args.kwonlyargs) == 0
                # TODO: check that no admin.display decorator already applied
            ):
                funcs[subnode.name] = FunctionDetails(subnode, "display")
        elif (
            isinstance(subnode, ast.Assign)
            and len(subnode.targets) == 1
            and isinstance((target := subnode.targets[0]), ast.Attribute)
            and isinstance(target.value, ast.Name)
            and target.value.id in funcs
            and target.attr in NAME_MAPS[funcs[target.value.id].decorator]
        ):
            names = NAME_MAPS[funcs[target.value.id].decorator]
            new_name = names[target.attr]
            funcs[target.value.id].assignments[new_name] = subnode

    if not admin_imported:
        return

    for name, funcdetails in funcs.items():
        if funcdetails.assignments:
            if funcdetails.node.decorator_list:
                offset = ast_start_offset(funcdetails.node.decorator_list[0])
                decorated = True
            else:
                offset = ast_start_offset(funcdetails.node)
                decorated = False
            yield (
                offset,
                partial(
                    decorate_function, funcdetails=funcdetails, decorated=decorated
                ),
            )
            for name, assignnode in funcdetails.assignments.items():
                yield ast_start_offset(assignnode), partial(erase_node, node=assignnode)
                yield (
                    ast_start_offset(assignnode.value),
                    partial(
                        store_value_src,
                        node=assignnode.value,
                        name=name,
                        funcdetails=funcdetails,
                    ),
                )


def decorate_function(
    tokens: list[Token], i: int, *, funcdetails: FunctionDetails, decorated: bool
) -> None:
    if decorated:
        i = reverse_find(tokens, i, name=OP, src="@")
    j, indent = extract_indent(tokens, i)
    dec_src = f"{indent}@admin.{funcdetails.decorator}(\n"

    # Pull args in predefined order
    names = NAME_MAPS[funcdetails.decorator]
    args = [
        (name, funcdetails.values[name])
        for name in names.values()
        if name in funcdetails.values
    ]
    comma = "," if len(args) > 1 else ""
    for name, source in args:
        source = source.replace("\n", f"\n{indent}    ")
        dec_src += f"{indent}    {name}={source}{comma}\n"
    dec_src += f"{indent})\n"
    insert(tokens, j, new_src=dec_src)


def store_value_src(
    tokens: list[Token],
    i: int,
    *,
    node: ast.expr,
    name: str,
    funcdetails: FunctionDetails,
) -> None:
    j = find_last_token(tokens, i, node=node)
    funcdetails.values[name] = tokens_to_src(tokens[i : j + 1])



"""
Drop branches for old Django versions like:
if django.VERSION >= (1, 8):
    ...
else:
    ...
"""

from __future__ import annotations

import ast
from collections.abc import Iterable
from functools import partial
from typing import Literal

from tokenize_rt import Offset, Token

from django_upgrade.ast import ast_start_offset, is_passing_comparison
from django_upgrade.data import Fixer, State, TokenFunc
from django_upgrade.tokens import Block

fixer = Fixer(
    __name__,
    min_version=(0, 0),
)


@fixer.register(ast.If)
def visit_If(
    state: State,
    node: ast.If,
    parents: tuple[ast.AST, ...],
) -> Iterable[tuple[Offset, TokenFunc]]:
    if (
        isinstance(node.test, ast.Compare)
        and (pass_fail := is_passing_comparison(node.test, state)) is not None
        and (
            # do not handle 'if ... elif ...'
            not node.orelse or not isinstance(node.orelse[0], ast.If)
        )
    ):
        yield (
            ast_start_offset(node),
            partial(
                _fix_block,
                node=node,
                keep_branch=("first" if pass_fail == "pass" else "second"),
            ),
        )


def _fix_block(
    tokens: list[Token],
    i: int,
    *,
    node: ast.If,
    keep_branch: Literal["first", "second"],
) -> None:
    if tokens[i].src != "if":
        # do not handle 'elif'
        return

    if node.orelse:
        if_block, else_block = _find_if_else_block(tokens, i)
        if keep_branch == "first":
            if_block.dedent(tokens)
            del tokens[if_block.end : else_block.end]
            del tokens[if_block.start : if_block.block]
        else:
            else_block.dedent(tokens)
            del tokens[if_block.start : else_block.block]
    else:
        if_block = Block.find(tokens, i, trim_end=True)
        if keep_branch == "first":
            if_block.dedent(tokens)
            del tokens[if_block.start : if_block.block]
        else:
            del tokens[if_block.start : if_block.end]


def _find_if_else_block(tokens: list[Token], i: int) -> tuple[Block, Block]:
    if_block = Block.find(tokens, i)
    i = if_block.end
    while tokens[i].src != "else":  # pragma: no cover
        i += 1
    else_block = Block.find(tokens, i, trim_end=True)
    return if_block, else_block
